// FlipFlat based on the Altinak Flip-Flat commands. Makes it possible to use it with ASCOM and INDI drivers

// Enable DEBUG in FlipFlat.h to see logging.
// Due to driver imitations in de indi driver, when using ESP32 Serial can't be used for FlipFlat communication.
// Use Serial 2 (pins 16 and 17) instead. Standard Serial is used for debug messages

#include <FlipFlat.h>

int deviceId = FLIP_FLAT; // set this to FLAT_MAN if you want to remove or not use the motor handling
int motorStatus = STOPPED;
int lightStatus = OFF;
int coverStatus = CLOSED;
int motorDirection = IDLE;
uint32_t steps = STEPS / 360.0 * COVER_ANGLE * MICROSTEPS;
int brightness = 0;
int lastCommand = CLOSE;

EasyButton coverButton(COVER_FF);

#if UNIPOLAR
AccelStepper stepper(AccelStepper::FULL4WIRE, STEP_PIN1, STEP_PIN3, STEP_PIN2, STEP_PIN4);
#else
AccelStepper stepper(AccelStepper::DRIVER, STEP_PIN, DIR_PIN);
#endif

TMC2130Stepper driver = TMC2130Stepper(CS_PIN, R_SENSE);

void setup()
{
  setupSerial();

  LOG("Debug mode");
  LOG("Starting FlipFlat");

  LOG("Setup open-close button");
  coverButton.begin();
  coverButton.onPressed(handleCoverButton);

  LOG("Initializing stepper");
  initializeStepper();
}

void loop()
{
  coverButton.read();

  if (FF_SERIAL.available() > 0) // all incoming communications are fixed length at 6 bytes including the \r
  {
    LOG("Command received");
    parseCommand();
  }

  if(!stepper.run())
    sendCommandResponse();

}

void initializeStepper()
{
  SPI.begin();

  pinMode(CS_PIN, OUTPUT);
  digitalWrite(CS_PIN, HIGH);

  driver.begin();          // Initiate pins and registeries
  driver.rms_current(600); // Set stepper current to 600mA. The command is the same as command TMC2130.setCurrent(600, 0.11, 0.5);
  driver.en_pwm_mode(1);   // Enable extremely quiet stepping
  driver.pwm_autoscale(1);
  driver.microsteps(MICROSTEPS);

  stepper.setMaxSpeed(MAX_SPEED);
  stepper.setAcceleration(1000);
  stepper.setEnablePin(EN_PIN);
  stepper.setPinsInverted(false, false, true);
  stepper.enableOutputs();
}

void parseCommand()
{
  LOG("Parse command");

  char *cmd;
  char *data;

  char str[20];
  memset(str, 0, 20);

  FF_SERIAL.readBytesUntil('\n', str, 20);

  cmd = str + 1;
  data = str + 2;

  // useful for debugging to make sure your commands came through and are parsed correctly.
  LOG1("cmd = ", cmd);
  LOG1("data = ", data);

  processCommand(cmd, data);

  while (FF_SERIAL.available() > 0)
    FF_SERIAL.read();
}

void processCommand(const char *cmd, const char *data)
{
  switch (*cmd)
  {
  /*
  Ping device
    Request: >POOO\r
    Return : *PiiOOO\n
      ii = DEVICE_ID
  */
  case 'P':
    LOG("Ping received");
    lastCommand = PING;
    break;

  /*
  Open shutter
  Request: >OOOO\r
  Return : *OiiOOO\n
    ii = deviceId
  This command is only supported on the Flip-Flat!
  */
  case 'O':
    LOG("Open received");
    lastCommand = OPEN;
    openFlipFlat();
    break;

  /*
  Close shutter
  Request: >COOO\r
  Return : *CiiOOO\n
    ii = deviceId
  This command is only supported on the Flip-Flat!
  */
  case 'C':
    LOG("Close received");
    lastCommand = CLOSE;
    closeFlipFlat();
    break;

  /*
  Turn light on
    Request: >LOOO\r
    Return : *LiiOOO\n
      ii = deviceId
  */
  case 'L':
    LOG("Light on received");
    lightStatus = ON;
    lastCommand = LIGHT_ON;
    // setBrightness(brightness);
    break;

  /*
  Turn light off
    Request: >DOOO\r
    Return : *DiiOOO\n
      id = deviceId
  */
  case 'D':
    LOG("Turn off received");
    lightStatus = OFF;
    lastCommand = LIGHT_OFF;
    // setBrightness(0);
    break;

  /*
  Set brightness
    Request: >Bxxx\r
      xxx = brightness value from 000-255
    Return : *Bidyyy\n
      id = deviceId
      yyy = value that brightness was set from 000-255
  */
  case 'B':
    LOG("Set brightness received");
    brightness = atoi(data);
    lastCommand = SET_BRIGHTNESS;
    // setBrightness(brightness);
    break;

  /*
  Get brightness
    Request: >J000\r
    Return : *Jiiyyy\n
      id = deviceId
      yyy = current brightness value from 000-255
  */
  case 'J':
    LOG("Get brightness received");
    lastCommand = GET_BRIGHTNESS;
    break;

  /*
  Get device status:
    Request: >S000\r
    Return : *SiiMLC\n
      ii = deviceId
      M  = motor status( 0 stopped, 1 running)
      L  = light status( 0 off, 1 on)
      C  = Cover Status( 0 moving, 1 closed, 2 open)
  */
  case 'S':
    LOG("Get status recived");
    lastCommand = STATUS;
    break;

  /*
  Get firmware version
    Request: >V000\r
    Return : *Vii001\n
      ii = deviceId
  */
  case 'V': // get firmware version
    LOG("Get firmware version received");
    lastCommand = VERSION;
    break;
  }
}

void openFlipFlat()
{
  if (coverStatus != OPENED)
  {
    LOG("Turn light off");
    // setBrightness(0);

    motorDirection = OPENING;
    rotateMotor(motorDirection);
  }
}


void closeFlipFlat()
{
  LOG("Closing FlipFlat");
  if (coverStatus == OPENED)
  {
    motorDirection = CLOSING;
    rotateMotor(motorDirection);
  }
  else
  {
    LOG("FlipFlap not opened");
  }
}


void rotateMotor(int direction)
{
  if (stepper.distanceToGo() == 0)
  {
    motorStatus = RUNNING;
    coverStatus = NEITHER_OPENED_NOR_CLOSED;

    stepper.move(direction * steps);
  }

  motorStatus = STOPPED;
  motorDirection = IDLE;
}

void setupSerial()
{

#ifdef ESP32
  FF_SERIAL.begin(9600, SERIAL_8N1, 16, 17);
  LOG_SERIAL.begin(9600);
#endif

#ifdef __AVR_ATmega32U4__
  FF_SERIAL.begin(9600);
#if DEBUG
  LOG_SERIAL.begin(9600);
#endif
#endif

  LOG("Logging serial up and running");
  LOG("FlipFlat Serial up and running");
}

void handleCoverButton()
{
  switch (coverStatus)
  {
  case CLOSED:
    LOG("Opening cover");
    openFlipFlat();
    break;
  case OPENED:
    LOG("Closing cover");
    closeFlipFlat();
    break;
  default:
    LOG("Cover not open or closed");
  }
}

void sendCommandResponse()
{
  switch (lastCommand)
  {
  case PING:
    FF_SERIAL << F("*P") << deviceId << F("OOO\n");
    lastCommand = NONE;
    break;
    
  case OPEN:
    //if (stepper.distanceToGo() == 0)
    //{
      coverStatus = OPENED;
      FF_SERIAL << F("*O") << deviceId << F("OOO\n");
      lastCommand = NONE;
    //}
    break;

  case CLOSE:
    //if (stepper.distanceToGo() == 0)
    //{
      coverStatus = CLOSED;
      FF_SERIAL << F("*C") << deviceId << F("OOO\n");
      lastCommand = NONE;
    //}
    break;

  case LIGHT_ON:
    FF_SERIAL << F("*L") << deviceId << F("OOO\n");
    lastCommand = NONE;
    break;

  case LIGHT_OFF:
    FF_SERIAL << F("*D") << deviceId << F("OOO\n");
    lastCommand = NONE;
    break;

  case SET_BRIGHTNESS:
    FF_SERIAL << F("*B") << deviceId << _WIDTHZ(brightness, 3) << F("\n");
    lastCommand = NONE;
    break;
    
  case GET_BRIGHTNESS:
    FF_SERIAL << F("*J") << deviceId << _WIDTHZ(brightness, 3) << F("\n");
    lastCommand = NONE;
    break;
        
  case STATUS:
    FF_SERIAL << F("*S") << deviceId << motorStatus << lightStatus << coverStatus << F("\n");
    lastCommand = NONE;
    break;
              
  case VERSION:
    FF_SERIAL << F("*V") << deviceId << F("AVG") << F("\n");
    lastCommand = NONE;
    break;

  default:
    lastCommand = NONE;
    break;
  }

}
